from swaystatus.click_event import ClickEvent

fake_click_event = ClickEvent(
    x=0,
    y=0,
    button=0,
    event=0,
    relative_x=0,
    relative_y=0,
    width=0,
    height=0,
    scale=0.0,
)



"""Generate a status line for swaybar."""

import argparse
from pathlib import Path

from . import __version__
from .app import App
from .config import Config
from .daemon import Daemon
from .env import environ_path, environ_paths, self_name
from .logging import configure_logging, logger
from .xdg import config_home, data_home


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=lambda prog: argparse.RawDescriptionHelpFormatter(
            prog, indent_increment=4, max_help_position=45
        ),
        epilog="See `pydoc swaystatus` for full documentation.",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=__version__,
    )
    parser.add_argument(
        "-c",
        "--config-file",
        metavar="FILE",
        type=Path,
        help="override configuration file",
    )
    parser.add_argument(
        "-C",
        "--config-dir",
        metavar="DIRECTORY",
        type=Path,
        help="override configuration directory",
    )
    parser.add_argument(
        "-D",
        "--data-dir",
        metavar="DIRECTORY",
        type=Path,
        help="override data directory",
    )
    parser.add_argument(
        "-I",
        "--include",
        metavar="DIRECTORY",
        type=Path,
        action="append",
        default=[],
        help="include an additional element package",
    )
    parser.add_argument(
        "-i",
        "--interval",
        metavar="SECONDS",
        type=float,
        help="override default update interval",
    )
    parser.add_argument(
        "--click-events",
        action="store_true",
        help="enable click events",
    )
    parser.add_argument(
        "--log-level",
        metavar="LEVEL",
        default="warning",
        choices=["debug", "info", "warning", "error", "critical"],
        help="override default minimum logging level (default: %(default)s)",
    )
    parser.add_argument(
        "order",
        metavar="NAME[:INSTANCE]",
        nargs="*",
        help="override configured element order",
    )
    return parser.parse_args()


def load_config(args: argparse.Namespace) -> Config:
    config_dir = environ_path("SWAYSTATUS_CONFIG_DIR") or args.config_dir or (config_home / self_name)
    config_file = environ_path("SWAYSTATUS_CONFIG_FILE") or args.config_file or (config_dir / "config.toml")
    config = Config.from_file(config_file) if config_file.is_file() else Config()
    data_dir = environ_path("SWAYSTATUS_DATA_DIR") or (data_home / self_name)
    package_path = environ_paths("SWAYSTATUS_PACKAGE_PATH") + [data_dir / "modules"]
    config.include = args.include + config.include + package_path
    if args.order:
        config.order = args.order
    if args.interval:
        config.interval = args.interval
    if args.click_events:
        config.click_events = True
    return config


def main() -> None:
    args = parse_args()
    config = load_config(args)
    configure_logging(args.log_level)
    daemon = Daemon(
        config.elements,
        config.interval,
        config.click_events,
    )
    try:
        App(daemon).run()
    except Exception:
        logger.exception("unhandled exception in app")


__all__ = [main.__name__]



from logging import Formatter, StreamHandler, basicConfig, getLogger

from .env import self_name

logger = getLogger(self_name)
log_format = "%(name)s: %(levelname)s: %(message)s"


def configure_logging(level_name: str) -> None:
    handler = StreamHandler()
    handler.setFormatter(Formatter(log_format))
    basicConfig(level=level_name.upper(), handlers=[handler])



"""
Locate and load content generation elements stored in modules.

A status bar element (a subclass of `swaystatus.BaseElement`) must be named
`Element` and must be stored in a module file at the top level of a package
that is visible in any or all of the following places (in order of preference):

    1. `--include=<DIRECTORY>` (can be used multiple times)

    2. Included in the configuration file:

        include = ['/path/to/package1', '/path/to/package2']

    3. Python packages in a colon-separated environment variable:

        SWAYSTATUS_PACKAGE_PATH="/path/to/package1:/path/to/package2"

    4. A python package in the data directory (in order of preference):

          a. `--data-dir=<DIRECTORY>`
          b. `$SWAYSTATUS_DATA_DIR/modules`
          c. `$XDG_DATA_HOME/swaystatus/modules`

    5. A python package with an entry point for `swaystatus.modules` defined
       like the following in the `pyproject.toml`:

          [project.entry-points."swaystatus.modules"]
          package = "awesome_swaystatus_modules"

When a module is found and imported, the the `Element.name` class attribute is
assigned the name used to look it up. This is done so that the class can be
located when delegating click events.

See the documentation for `swaystatus.element` to learn about creating elements.
"""

import sys
from functools import cache, cached_property
from importlib import import_module, metadata
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from typing import Iterable
from uuid import uuid4

from .element import BaseElement
from .logging import logger


class ModuleRegistry:
    """Track, locate, and import status bar elements."""

    def __init__(self, include: Iterable[str | Path]) -> None:
        self.include = [Path(i).expanduser() for i in include or []]

    @cached_property
    def packages(self) -> list[str]:
        """Return recognized packages names in order of preference."""
        result = []
        for package_dir in self.include:
            if (init_file := package_dir / "__init__.py").is_file():
                package_name = str(uuid4()).replace("-", "")
                if spec := spec_from_file_location(package_name, init_file):
                    package = module_from_spec(spec)
                    sys.modules[package_name] = package
                    if spec.loader:
                        spec.loader.exec_module(package)
                        result.append(package_name)
        for entry_point in metadata.entry_points(group="swaystatus.modules"):
            result.append(entry_point.load().__name__)
        return result

    @cache
    def element_class(self, name: str) -> type[BaseElement]:
        """Return the first matching element constructor in any visible packages."""
        for package in self.packages:
            try:
                module = import_module(f"{package}.{name}")
                if hasattr(module, "Element") and issubclass(module.Element, BaseElement):
                    logger.debug(f"imported module: {module!r}")
                    module.Element.name = name
                    return module.Element
            except ModuleNotFoundError:
                pass
        else:
            raise ModuleNotFoundError(f"Module not found in any package: {name}")


__all__ = [ModuleRegistry.__name__]



from dataclasses import replace
from pathlib import Path
from random import randint
from subprocess import Popen

import pytest

from swaystatus import BaseElement, ClickEvent
from swaystatus.element import ClickHandlerResult, ShellCommand

from .fake import fake_click_event


def test_base_element_blocks_not_implemented() -> None:
    """Ensure that the block generator is implemented on subclasses."""
    with pytest.raises(NotImplementedError):
        BaseElement().blocks()


def test_element_on_click_method() -> None:
    """Ensure that handlers can be defined as a method."""
    actual_button: int | None = None
    expected_button = randint(1, 3)

    class Element(BaseElement):
        def register_click(self, button: int) -> None:
            nonlocal actual_button
            actual_button = button

        def on_click_1(self, click_event: ClickEvent):
            self.register_click(1)

        def on_click_2(self, click_event: ClickEvent):
            self.register_click(2)

        def on_click_3(self, click_event: ClickEvent):
            self.register_click(3)

    element = Element()
    element.on_click(replace(fake_click_event, button=expected_button))
    assert expected_button == actual_button


def test_element_on_click_function() -> None:
    """Ensure that function handlers can be set at initialization."""

    class Element(BaseElement):
        name = "test"

    clicked_element: BaseElement | None = None

    def handler(element: Element, click_event: ClickEvent):
        nonlocal clicked_element
        clicked_element = element

    button = randint(1, 5)
    element = Element(on_click={button: handler})
    element.on_click(replace(fake_click_event, button=button))
    assert clicked_element is element


def test_element_on_click_shell_command(tmp_path) -> None:
    """Ensure that shell command handlers can be set at initialization."""

    class Element(BaseElement):
        name = "test"

    button = randint(1, 5)
    cases = {
        "$foo": "some string",  # environment variables added
        "${button}": str(button),  # environment variables from click event
        "~": str(Path.home()),  # shell tilde expansion
    }
    env = {"foo": cases["$foo"]}
    click_event = replace(fake_click_event, button=button)
    stdout_file = tmp_path / "stdout"
    for s, expected_output in cases.items():
        handler = f"echo {s} >{stdout_file}"  # shell redirection
        element = Element(on_click={button: handler}, env=env)
        process = element.on_click(click_event)
        assert isinstance(process, Popen)
        process.wait()
        actual_output = stdout_file.read_text().strip()
        assert actual_output == expected_output


def test_element_on_click_function_return_passthrough() -> None:
    """Ensure that a function handler's return value is preserved sometimes."""

    class Element(BaseElement):
        name = "test"

    for expected_value in (None, False, True, lambda: True, Popen("true")):

        def handler(element: Element, click_event: ClickEvent) -> ClickHandlerResult:
            return expected_value

        button = randint(1, 5)
        element = Element(on_click={button: handler})
        actual_value = element.on_click(replace(fake_click_event, button=button))
        assert actual_value is expected_value


def test_element_on_click_function_return_shell_command_run() -> None:
    """Ensure that a function handler's return value is run if it's a shell command."""

    class Element(BaseElement):
        name = "test"

    for expected_args in ("true", ["true"]):

        def handler(element: Element, click_event: ClickEvent, /) -> ShellCommand | ClickHandlerResult:
            return expected_args

        button = randint(1, 5)
        element = Element(on_click={button: handler})
        process = element.on_click(replace(fake_click_event, button=button))
        assert isinstance(process, Popen)
        assert process.args is expected_args



"""
Configuring swaystatus to do your bidding.

Configuration is defined in a toml file located in one of the following places
(in order of preference):

    1. `--config-file=<FILE>`
    2. `$SWAYSTATUS_CONFIG_FILE`
    3. `<DIRECTORY>/config.toml` where `<DIRECTORY>` is from `--config-dir=<DIRECTORY>`
    4. `$SWAYSTATUS_CONFIG_DIR/config.toml`
    5. `$XDG_CONFIG_HOME/swaystatus/config.toml`
    6. `$HOME/.config/swaystatus/config.toml`

The following keys are recognized in the configuration file:

    `order` (type: `list[str]`, default: `[]`)
        The desired elements to display and their order. Each item can be of
        the form "name" or "name:instance". The latter form allows the same
        element to be used multiple times with different settings for each
        "instance".

    `interval` (type: `float`, default: `5.0`)
        How often (in seconds) to update the status bar.

    `click_events` (type: `bool`, default: `false`)
        Whether or not to listen for status bar clicks.

    `include` (type: `list[str]`, default: `[]`)
        Additional directories to treat as element packages.

    `env` (type: `dict[str, str]`, default: `{}`)
        Additional environment variables visible to click handlers.

    `on_click` (type: `dict[int, str | list[str]]`, default: `{}`)
        Maps pointer button numbers to shell commands that should be run in
        response to a click by that button.

    `settings` (type: `dict[str, dict[str, Any]]`, default: `{}`)
        Maps element specifiers (as defined in `order`) to keyword arguments
        that will be passed to the element constructor.

A typical configuration file might look like the following:

    order = [
        'hostname',
        'path_exists:/mnt/foo',
        'memory',
        'clock',
        'clock:home'
    ]

    click_events = true

    [env]
    terminal = 'foot'

    [settings.hostname]
    full_text = "host: {}"
    on_click.1 = '$terminal --hold hostnamectl'

    [settings.path_exists]
    on_click.1 = ['$terminal', '--working-directory=$instance']
    on_click.2 = ['$terminal', '--hold', 'df', '$instance']

    [settings.clock]
    on_click.1 = '$terminal --hold cal'

    [settings."clock:home".env]
    TZ = 'Asia/Tokyo'
"""

import tomllib
from dataclasses import dataclass, field
from functools import cache, cached_property
from pathlib import Path
from typing import Any, Self

from .element import BaseElement
from .logging import logger
from .modules import ModuleRegistry

default_interval = 5.0


@dataclass(kw_only=True, eq=False)
class Config:
    """Data class representing runtime configuration."""

    order: list[str] = field(default_factory=list)
    interval: float = default_interval
    click_events: bool = False
    settings: dict[str, Any] = field(default_factory=dict)
    include: list[str | Path] = field(default_factory=list)
    on_click: dict[int, str | list[str]] = field(default_factory=dict)
    env: dict[str, str] = field(default_factory=dict)

    @cached_property
    def module_registry(self) -> ModuleRegistry:
        return ModuleRegistry(self.include)

    @cache
    def element(self, key: str) -> BaseElement:
        """Return an element instance for a configuration key."""
        logger.info(f"loading element {key=!r}")
        name, instance = decode_element_key(key)
        Element = self.module_registry.element_class(name)
        kwargs = self.settings.get(name, {})
        if instance:
            kwargs = deep_merge_dicts(kwargs, self.settings.get(key, {}))
        kwargs["env"] = self.env | kwargs.get("env", {})
        kwargs["instance"] = instance
        logger.debug(f"initializing element {name=!r}: {kwargs=!r}")
        return Element(**kwargs)

    @cached_property
    def elements(self) -> list[BaseElement]:
        """Return all status bar content producers in order."""
        return list(map(self.element, self.order))

    @classmethod
    def from_file(cls, file: Path) -> Self:
        """Instantiate a configuration object from a toml file."""
        return cls(**tomllib.load(file.open("rb")))


def decode_element_key(key: str) -> tuple[str, str | None]:
    """Parse an element name and instance from a string like "name" or "name:instance"."""
    try:
        name, instance = key.split(":", maxsplit=1)
    except ValueError:
        name, instance = key, None
    assert name
    return name, instance


def deep_merge_dicts(first: dict, second: dict) -> dict:
    """Recursively merge the second dictionary into the first."""
    result = first.copy()
    for key, value in second.items():
        if key in result and isinstance(result[key], dict) and isinstance(value, dict):
            result[key] = deep_merge_dicts(result[key], value)
        else:
            result[key] = value
    return result


__all__ = [Config.__name__]



from typing import Iterable

from .element import BaseElement
from .input import InputProcessor
from .logging import logger
from .output import OutputProcessor
from .threading import InputReader, OutputWriter


class Daemon:
    """Coordinator of input and output."""

    def __init__(self, elements: Iterable[BaseElement], interval: float, click_events: bool) -> None:
        self.output_writer = OutputWriter(OutputProcessor(elements, click_events=click_events), interval)
        self.input_reader = InputReader(InputProcessor(elements), self.output_writer) if click_events else None

    def update(self) -> None:
        self.output_writer.update()

    def stop(self) -> None:
        logger.info("stopping daemon")
        self.output_writer.stop()

    def start(self) -> None:
        logger.info("starting daemon")
        if self.input_reader:
            self.input_reader.start()
        self.output_writer.start()


__all__ = [Daemon.__name__]



from signal import SIGCONT, SIGINT, SIGTERM, SIGUSR1, Signals, signal
from types import FrameType

from .daemon import Daemon
from .logging import logger


class App:
    """Manager for the daemon's lifecycle."""

    def __init__(self, daemon: Daemon) -> None:
        self.daemon = daemon

    def update(self, sig: int, frame: FrameType | None) -> None:
        logger.info(f"signaled to update: {Signals(sig).name} ({sig})")
        logger.debug(f"current stack frame: {frame!r}")
        self.daemon.update()

    def shutdown(self, sig: int, frame: FrameType | None) -> None:
        logger.info(f"signaled to shutdown: {Signals(sig).name} ({sig})")
        logger.debug(f"current stack frame: {frame!r}")
        self.daemon.stop()

    def run(self) -> None:
        signal(SIGUSR1, self.update)
        signal(SIGCONT, self.update)
        signal(SIGINT, self.shutdown)
        signal(SIGTERM, self.shutdown)
        self.daemon.start()


__all__ = [App.__name__]



import pytest

from swaystatus.config import Config, decode_element_key
from swaystatus.element import BaseElement
from swaystatus.modules import ModuleRegistry


def test_decode_element_key() -> None:
    """Ensure that element key is parsed."""
    assert decode_element_key("foo") == ("foo", None)
    assert decode_element_key("foo:") == ("foo", "")
    assert decode_element_key("foo:a") == ("foo", "a")
    assert decode_element_key("foo:a:b") == ("foo", "a:b")


def test_decode_element_key_no_name() -> None:
    """Ensure that an element key contains the name."""
    with pytest.raises(AssertionError):
        decode_element_key("")
        decode_element_key(":")
        decode_element_key(":a")
        decode_element_key(":a:b")


def test_config_element_singleton(monkeypatch) -> None:
    """Ensure that elements with the same key map to the same instance."""

    # For this test, it's only important that the element class returned is
    # valid, not unique to the name. The only question is whether or not the
    # same instance is returned for the same key.

    class Element(BaseElement):
        pass

    def element_class(self, name: str) -> type[BaseElement]:
        return Element

    monkeypatch.setattr(ModuleRegistry, "element_class", element_class)

    config = Config()

    assert config.element("foo") is config.element("foo")
    assert config.element("foo") is not config.element("bar")
    assert config.element("foo") is not config.element("foo:a")
    assert config.element("foo:a") is config.element("foo:a")
    assert config.element("foo:a") is not config.element("foo:b")



from subprocess import Popen
from typing import Callable

from .click_event import ClickEvent

type ShellCommand = str | list[str]
type ClickHandlerResult = Popen | Callable[[], bool] | bool | None
type ClickHandler[E] = ShellCommand | Callable[[E, ClickEvent], ShellCommand | ClickHandlerResult]



import json
import random
from io import StringIO
from signal import SIGCONT, SIGSTOP
from typing import Iterator

from swaystatus import BaseElement, Block
from swaystatus.output import OutputProcessor


def test_output_blocks_multiple_elements() -> None:
    """Ensure that multiple elements output their blocks in the correct order."""

    class Element1(BaseElement):
        name = "test1"

        def blocks(self) -> Iterator[Block]:
            yield self.block("foo")

    class Element2(BaseElement):
        name = "test2"

        def blocks(self) -> Iterator[Block]:
            yield self.block("bar")

    output_processor = OutputProcessor([Element1(), Element2()])
    actual_blocks = list(output_processor.blocks())
    expected_blocks = [
        Block(name="test1", full_text="foo"),
        Block(name="test2", full_text="bar"),
    ]
    assert actual_blocks == expected_blocks


def test_output_blocks_multiple_blocks_per_element() -> None:
    """Ensure that a single element is able to output multiple blocks."""
    texts = ["foo", "bar", "baz"]

    class Element(BaseElement):
        name = "test"

        def blocks(self) -> Iterator[Block]:
            for text in texts:
                yield self.block(text)

    output_processor = OutputProcessor([Element()])
    actual_blocks = list(output_processor.blocks())
    expected_blocks = [Block(name="test", full_text=text) for text in texts]
    assert actual_blocks == expected_blocks


def test_output_blocks_identical_elements_cached() -> None:
    """Ensure that identical elements are only polled once per iteration."""

    class Element(BaseElement):
        name = "test"
        count = 0

        def blocks(self) -> Iterator[Block]:
            self.count += 1
            yield self.block(str(self.count))

    element = Element()

    # Show that separate polls produce different blocks.
    assert list(element.blocks()) != list(element.blocks())

    output_processor = OutputProcessor([element, element])

    # Show that the blocks produced by the element are reused.
    blocks_first = list(output_processor.blocks())
    assert blocks_first[0] == blocks_first[1]

    # Show that the blocks produced by the last call are not reused.
    blocks_second = list(output_processor.blocks())
    assert blocks_first != blocks_second


def test_output_header_click_events() -> None:
    """Ensure that the click events setting is reflected in output header."""

    for click_events in [None, False, True]:
        kwargs = {}
        if click_events is not None:
            kwargs["click_events"] = click_events
        output_processor = OutputProcessor([], click_events=click_events)
        assert output_processor.header["click_events"] == bool(click_events)


def test_output_process_encoded() -> None:
    """Ensure that output is written in the expected format."""

    class Element(BaseElement):
        name = "test"
        count = 0

        def blocks(self) -> Iterator[Block]:
            self.count += 1
            yield self.block(f"iteration {self.count}")

    output_file = StringIO()
    output_processor = OutputProcessor([Element()])
    status = output_processor.process(output_file)

    def next_iteration() -> tuple[list[Block], list[str]]:
        pos = output_file.tell()
        blocks = next(status)
        output_file.seek(pos)
        lines = output_file.readlines()
        return blocks, lines

    def block(i: int) -> Block:
        return Block(full_text=f"iteration {i}", name="test")

    def body_line(i: int) -> str:
        return f",[{json.dumps(dict(full_text=f'iteration {i}', name='test'))}]\n"

    blocks, lines = next_iteration()
    assert len(lines) == 3
    assert blocks == [block(1)]
    header = json.loads(lines[0])
    assert header["version"] == 1
    assert header["stop_signal"] == SIGSTOP
    assert header["cont_signal"] == SIGCONT
    assert not header["click_events"]
    assert lines[1] == "[[]\n"  # start of infinite body array
    assert lines[2] == body_line(1)
    for i in range(2, random.randint(5, 10)):
        blocks, lines = next_iteration()
        assert blocks == [block(i)]
        assert lines == [body_line(i)]



from functools import cache, cached_property
from json import JSONEncoder
from signal import SIGCONT, SIGSTOP
from typing import IO, Any, Iterable, Iterator

from .block import Block
from .element import BaseElement
from .logging import logger


class OutputProcessor:
    """Assemble blocks generated by elements into a status line."""

    def __init__(self, elements: Iterable[BaseElement], click_events=False) -> None:
        self.elements = list(elements)
        self.click_events = bool(click_events)

    @cached_property
    def header(self) -> dict[str, Any]:
        return dict(
            version=1,
            stop_signal=SIGSTOP,
            cont_signal=SIGCONT,
            click_events=self.click_events,
        )

    def blocks(self) -> Iterator[Block]:
        """Yield blocks from every element."""
        element_blocks.cache_clear()
        for element in self.elements:
            yield from element_blocks(element)

    def lines(self) -> Iterator[list[Block]]:
        while True:
            yield list(self.blocks())

    def process(self, file: IO[str]) -> Iterator[list[Block]]:
        """Send status lines to output and yield the originating blocks."""

        def send(line: str) -> None:
            print(line, file=file, flush=True)

        encoder = Encoder()
        send(encoder.encode(self.header))
        send("[[]")
        for line in self.lines():
            send(",{}".format(encoder.encode(line)))
            yield line


@cache
def element_blocks(element: BaseElement) -> list[Block]:
    try:
        return list(element.blocks())
    except Exception:
        logger.exception(f"exception while getting blocks for {element}")
        return []


class Encoder(JSONEncoder):
    def default(self, block: Block):
        return block.as_dict()


__all__ = [OutputProcessor.__name__]



from functools import cache, cached_property
from json import JSONDecoder
from typing import IO, Iterable, Iterator

from .click_event import ClickEvent
from .element import BaseElement, ClickHandlerResult
from .logging import logger


class InputProcessor:
    """Handle click events, sending them to the appropriate element's handler."""

    def __init__(self, elements: Iterable[BaseElement]) -> None:
        self.elements = list(elements)

    @cached_property
    def elements_by_key(self) -> dict[tuple[str | None, str | None], BaseElement]:
        """Provide fast lookup of elements by name and instance."""
        return {(e.name, e.instance): e for e in self.elements}

    @cache
    def element(self, name: str | None, instance: str | None) -> BaseElement:
        """
        Return the handler for element identifiers.

        Try to find an element matching the given name and instance.
        If a matching element is not found, look for one with the same name.
        Otherwise, raise `KeyError`.
        """
        for instance in (instance, None):
            try:
                return self.elements_by_key[(name, instance)]
            except KeyError:
                pass
        raise KeyError((name, instance))

    def process(self, file: IO[str]) -> Iterator[tuple[ClickEvent, BaseElement, ClickHandlerResult]]:
        """Yield click events and their corresponding element handler results."""
        for click_event in click_events(file):
            logger.debug(f"received click event: {click_event!r}")
            try:
                element = self.element(click_event.name, click_event.instance)
            except KeyError:
                logger.warn(f"no element to handle {click_event}")
                continue
            logger.info(f"sending {click_event} to {element}")
            handler_result = element.on_click(click_event)
            logger.debug(f"{element} handled {click_event} with result: {handler_result!r}")
            yield click_event, element, handler_result


def click_events(file: IO[str]) -> Iterator[ClickEvent]:
    """Yield decoded click events from a file."""
    decoder = Decoder()
    assert file.readline().strip() == "["
    for line in file:
        try:
            yield decoder.decode(line.strip().lstrip(","))
        except Exception:
            logger.exception(f"exception while decoding input: {line!r}")


class Decoder(JSONDecoder):
    def __init__(self) -> None:
        super().__init__(object_hook=lambda kwargs: ClickEvent(**kwargs))


__all__ = [InputProcessor.__name__]



"""
Framework for creating an interactive status line for swaybar.

Typical usage consists of the following:

    1. Create elements by subclassing `BaseElement` to produce blocks.
        - See `swaystatus.element` about creating elements.
        - See `swaystatus.modules` about where to put element modules.

    2. Configure swaystatus to use those elements.
        - See `swaystatus.config` about enabling and configuring elements.

    3. Produce content for swaybar with the `swaystatus` command.
        - See `swaystatus --help` for command line usage.
        - See `status_command` in sway-bar(5) to set the status command.

See swaybar-protocol(7) for a full description of the status bar protocol.
"""

import locale

from .block import Block
from .click_event import ClickEvent
from .element import BaseElement

locale.setlocale(locale.LC_ALL, "")

__version__ = "0.13.1"

__all__ = [
    Block.__name__,
    ClickEvent.__name__,
    BaseElement.__name__,
]



from subprocess import PIPE, Popen
from threading import Thread
from typing import Callable, Iterator

from .logging import logger
from .typing import ShellCommand


class ProxyThread[T](Thread):
    """Dedicated thread for handling items from a generator."""

    source: Iterator[T]
    handler: Callable[[T], None]

    def __init__(self, source: Iterator[T], handler: Callable[[T], None]) -> None:
        super().__init__()
        self.source = source
        self.handler = handler

    def run(self) -> None:
        for item in self.source:
            self.handler(item)


class ShellCommandProcess(Popen):
    """Run a shell command, logging stdout and stderr."""

    def __init__(self, command: ShellCommand) -> None:
        super().__init__(command, stdout=PIPE, stderr=PIPE, shell=True, text=True)
        assert self.stdout and self.stderr
        ProxyThread(self.stdout, logger.debug).start()
        ProxyThread(self.stderr, logger.error).start()



"""A click event describes a block that was clicked by a pointer."""

from dataclasses import asdict, dataclass
from typing import Any


@dataclass(slots=True, kw_only=True)
class ClickEvent:
    """Data class representing an event generated when clicking on a status bar block."""

    name: str | None = None
    instance: str | None = None
    x: int
    y: int
    button: int
    event: int
    relative_x: int
    relative_y: int
    width: int
    height: int
    scale: float

    def __str__(self) -> str:
        return f"click event button={self.button}"

    def as_dict(self) -> dict[str, Any]:
        """Return a dict representation of this instance without any unset values."""
        return {name: value for name, value in asdict(self).items() if value is not None}



import sys
from subprocess import Popen
from threading import Event, Thread

from .element import ClickHandlerResult
from .input import InputProcessor
from .logging import logger
from .output import OutputProcessor


class OutputWriter:
    """Write status lines to an output file at a regular interval and when requested."""

    file = sys.stdout

    def __init__(self, output_processor: OutputProcessor, interval: float) -> None:
        self.output_processor = output_processor
        self.interval = interval
        self._tick = Event()
        self._running = Event()

    def update(self) -> None:
        logger.info("updating status bar")
        self._tick.set()

    def stop(self) -> None:
        logger.info("stopping output")
        self._running.clear()
        self._tick.set()

    def start(self) -> None:
        logger.info("starting output")
        self._running.set()
        for blocks in self.output_processor.process(self.file):
            logger.debug(f"processed output: {blocks}")
            self._tick.clear()
            self._tick.wait(self.interval)
            if not self._running.is_set():
                break


class InputReader(Thread):
    """Send click events from an input file to a targed element and handle results."""

    daemon = True
    file = sys.stdin

    def __init__(self, input_processor: InputProcessor, output_writer: OutputWriter) -> None:
        super().__init__(name="input")
        self.input_processor = input_processor
        self.output_writer = output_writer

    def run(self) -> None:
        logger.info("starting input")
        for click_event, element, handler_result in self.input_processor.process(self.file):
            UpdateHandler(self.output_writer, handler_result).start()


class UpdateHandler(Thread):
    """Update the status line only after successful handler completion."""

    daemon = True

    def __init__(self, output_writer: OutputWriter, handler_result: ClickHandlerResult) -> None:
        super().__init__(name="update")
        self.output_writer = output_writer
        self.handler_result = handler_result

    def wait(self) -> bool:
        return (
            (isinstance(self.handler_result, Popen) and self.handler_result.wait() == 0)
            or (callable(self.handler_result) and self.handler_result())
            or bool(self.handler_result)
        )

    def run(self) -> None:
        if self.wait():
            self.output_writer.update()



import os
import sys
from contextlib import contextmanager
from pathlib import Path

self_name = os.path.basename(sys.argv[0])


def environ_path(name: str) -> Path | None:
    """Return a path from an environment variable (if set)."""
    if value := os.environ.get(name):
        return Path(value).expanduser()
    return None


def environ_paths(name: str) -> list[Path]:
    """Return a list of paths from and environment variable."""
    return [Path(p).expanduser() for p in os.environ[name].split(":")] if name in os.environ else []


@contextmanager
def environ_update(**kwargs):
    """Alter the environment during execution of a block."""
    environ_save = os.environ.copy()
    os.environ.update({k: str(v) for k, v in kwargs.items()})
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(environ_save)



import importlib
import shutil
import sys
from pathlib import Path

import pytest

from swaystatus.modules import ModuleRegistry


@pytest.fixture
def temp_module(tmp_path):
    def copy(src_name: str | None = None, dst_name: str | None = None) -> Path:
        src = Path(__file__).parent / "modules" / (src_name or "no_output.py")
        dst = tmp_path / (dst_name or src.name)
        dst.parent.mkdir(parents=True, exist_ok=True)
        (dst.parent / "__init__.py").touch()
        shutil.copyfile(src, dst)
        return dst

    return copy


def test_modules_load_module_not_found() -> None:
    """Ensure that requesting a non-existent module will raise an error."""
    with pytest.raises(ModuleNotFoundError, match="foo"):
        registry = ModuleRegistry([])
        registry.packages = []
        registry.element_class("foo")


def test_modules_load(temp_module) -> None:
    """Ensure that an existing module will be found in a valid package."""
    path = temp_module(dst_name="foo.py")
    modules = ModuleRegistry([path.parent])
    Element = modules.element_class("foo")
    assert sys.modules[Element.__module__].__file__ == str(path)


def test_modules_load_first_found(temp_module) -> None:
    """Ensure packages included earlier have preference when looking for a module."""
    name = "foo"
    path1 = temp_module(dst_name=f"a/{name}.py")
    path2 = temp_module(dst_name=f"b/{name}.py")
    registry = ModuleRegistry([path1.parent, path2.parent])
    Element = registry.element_class(name)
    assert sys.modules[Element.__module__].__file__ == str(path1)


def test_modules_entry_points(temp_module, monkeypatch) -> None:
    """Ensure that module packages defined as an entry point are recognized."""

    class Package:
        __name__ = "entry"

    class EntryPoint:
        def load(self):
            return Package()

    def entry_points(**kwargs):
        assert kwargs["group"] == "swaystatus.modules"
        return [EntryPoint()]

    assert hasattr(importlib, "metadata")
    monkeypatch.setattr(importlib.metadata, "entry_points", entry_points)
    registry = ModuleRegistry([temp_module().parent])
    assert len(registry.packages) == 2  # tmp_path and the fake entry point
    assert registry.packages[-1] == "entry"  # the fake entry point is after tmp_path



import dataclasses
import json
from io import StringIO
from operator import itemgetter
from random import shuffle
from subprocess import Popen
from typing import IO, Callable, Iterable

from swaystatus.click_event import ClickEvent
from swaystatus.element import BaseElement
from swaystatus.input import InputProcessor

from .fake import fake_click_event


def create_input_file(click_events: Iterable[ClickEvent]) -> IO[str]:
    input_file = StringIO()
    input_file.write("[\n")
    for click_event in click_events:
        input_file.write(f",{json.dumps(click_event.as_dict())}\n")
    input_file.seek(0)
    return input_file


def test_input_process_click_handler_delegation() -> None:
    """Ensure that clicks are sent to the correct handler."""
    actual_clicks = []

    def handler_func() -> bool:
        return True

    handler_process = Popen("true")

    class Element(BaseElement):
        name = "test"

        def register_click(self, click_event: ClickEvent) -> None:
            nonlocal actual_clicks
            actual_clicks.append(click_event)

        def on_click_1(self, click_event: ClickEvent) -> None:
            self.register_click(click_event)

        def on_click_2(self, click_event: ClickEvent) -> bool:
            self.register_click(click_event)
            return True

        def on_click_3(self, click_event: ClickEvent) -> Callable[[], bool]:
            self.register_click(click_event)
            return handler_func

        def on_click_4(self, click_event: ClickEvent) -> Popen:
            self.register_click(click_event)
            return handler_process

    element = Element()
    expected_processed = [
        (
            dataclasses.replace(
                fake_click_event,
                name="test",
                button=button,
            ),
            element,
            handler_result,
        )
        for button, handler_result in enumerate(
            [
                None,
                True,
                handler_func,
                handler_process,
            ],
            start=1,
        )
    ]
    shuffle(expected_processed)
    expected_clicks = list(map(itemgetter(0), expected_processed))
    input_file = create_input_file(expected_clicks)
    input_processor = InputProcessor([element])
    actual_processed = list(input_processor.process(input_file))
    assert actual_clicks == expected_clicks
    assert actual_processed == expected_processed


def test_input_process_element_delegation() -> None:
    """Ensure that clicks are sent to the correct element."""
    actual_clicks = []

    class Element(BaseElement):
        def on_click_1(self, click_event: ClickEvent) -> None:
            nonlocal actual_clicks
            actual_clicks.append(click_event)

    class Element1(Element):
        name = "test1"

    class Element2(Element):
        name = "test2"

    elements = [
        Element1(),
        Element1(instance="a"),
        Element1(instance="b"),
        Element2(),
        Element2(instance="a"),
        Element2(instance="b"),
    ]
    expected_processed = [
        (
            dataclasses.replace(
                fake_click_event,
                name=element.name,
                instance=element.instance,
                button=1,
            ),
            element,
            None,
        )
        for element in elements
    ]
    shuffle(expected_processed)
    expected_clicks = list(map(itemgetter(0), expected_processed))
    input_file = create_input_file(expected_clicks)
    input_processor = InputProcessor(elements)
    actual_processed = list(input_processor.process(input_file))
    assert actual_clicks == expected_clicks
    assert actual_processed == expected_processed



from random import shuffle
from typing import Any

from swaystatus import Block


def test_block_dict_minimal() -> None:
    """Ensure that only the keyword arguments passed are included in its `dict` form."""
    kwarg_pairs: list[tuple[str, Any]] = [
        ("full_text", "full"),
        ("short_text", "short"),
        ("color", "#eeeeee"),
        ("background", "#ffffff"),
        ("border", "#000000"),
        ("border_top", 1),
        ("border_bottom", 2),
        ("border_left", 3),
        ("border_right", 4),
        ("min_width", 100),
        ("align", "center"),
        ("name", "foo"),
        ("instance", "bar"),
        ("urgent", False),
        ("separator", ","),
        ("separator_block_width", 2),
        ("markup", "pango"),
    ]
    shuffle(kwarg_pairs)
    for i in range(len(kwarg_pairs)):
        kwargs = dict(kwarg_pairs[:i])
        assert Block(**kwargs).as_dict() == kwargs



"""A block is a unit of content for the status bar."""

from dataclasses import asdict, dataclass
from typing import Any


@dataclass(slots=True, kw_only=True)
class Block:
    """Data class representing a unit of status bar content."""

    full_text: str | None = None
    short_text: str | None = None
    color: str | None = None
    background: str | None = None
    border: str | None = None
    border_top: int | None = None
    border_bottom: int | None = None
    border_left: int | None = None
    border_right: int | None = None
    min_width: int | str | None = None
    align: str | None = None
    name: str | None = None
    instance: str | None = None
    urgent: bool | None = None
    separator: bool | None = None
    separator_block_width: int | None = None
    markup: str | None = None

    def __str__(self) -> str:
        return f"block full_text={self.full_text!r}"

    def as_dict(self) -> dict[str, Any]:
        """Return a dict representation of this instance without any unset values."""
        return {name: value for name, value in asdict(self).items() if value is not None}



import tomllib
from pathlib import Path

import swaystatus


def test_version_matches_project() -> None:
    """Ensure that the package version matches the project version."""
    project_file = Path(__file__).resolve().parents[1] / "pyproject.toml"
    project_version = tomllib.load(project_file.open("rb"))["project"]["version"]
    assert swaystatus.__version__ == project_version



"""
An element produces blocks of content to display in the status bar.

When the status bar is running, every configured element will be prompted for
blocks at the configured interval. Once every element has replied, all of the
collected blocks are encoded and sent to the bar via stdout.
"""

from subprocess import Popen
from types import MethodType
from typing import Iterator, Self

from .block import Block
from .click_event import ClickEvent
from .env import environ_update
from .logging import logger
from .subprocess import ShellCommandProcess
from .typing import ClickHandler, ClickHandlerResult, ShellCommand


class BaseElement:
    """
    Base class for constructing elements for the status bar.

    The subclass must be named `Element` and be contained in a module file
    whose name will be used by swaystatus to identify it. For example, if there
    is a module file named `clock.py`, the class will have an attribute `name`
    set to "clock".

    The `blocks` method is called to generate output at every interval and must
    be overridden. There is no requirement regarding the number of blocks that
    are yielded, however if no blocks are yielded, nothing will be visible in
    the status bar for that element. This may be desirable if, for example, the
    element yields a block for every mounted disk and there are none mounted.

    A hypothetical clock module file might contain the following:

        >>> from time import strftime
        >>> from typing import Iterator
        >>> from swaystatus import BaseElement, Block
        >>> class Element(BaseElement):
        >>>     def blocks(self) -> Iterator[Block]:
        >>>         yield self.block(strftime("%c"))

    Enable the element by adding it to the configuration file:

        order = ["clock"]

    Let's make the clock respond to a left mouse button click by running a
    shell command. Enable click events and add a click handler to the settings
    for that element:

        order = ["clock"]

        click_events = true

        [settings.clock]
        on_click.1 = "foot --hold cal"

    Maybe there needs to be an additional clock that always shows a specific
    timezone:

        order = ["clock", "clock:home"]

        click_events = true

        [settings.clock]
        on_click.1 = "foot --hold cal"

        [settings."clock:home".env]
        TZ = 'Asia/Tokyo'

    The "name:instance" form in `order` allows multiple instances of the same
    element, each having their own settings.

    Because the module is named `clock.py`, swaystatus will set the class
    attribute `name` to "clock" for the first element ("clock" in `order`) as
    if it had been declared like:

        >>> from swaystatus import BaseElement
        >>> class Element(BaseElement):
        >>>     name = "clock"

    Because the second element ("clock:home") is using the "name:instance"
    form, it will also have the instance attribute `instance` set to "home" as
    if it had been declared like:

        >>> from swaystatus import BaseElement
        >>> class Element(BaseElement):
        >>>     name = "clock"
        >>> element = Element(instance="home")

    You should not set these attributes yourself, as it will confuse and sadden
    swaystatus, and it will propably not respond to your clicks.
    """

    name: str

    def __init__(
        self,
        *,
        instance: str | None = None,
        env: dict[str, str] | None = None,
        on_click: dict[int, ClickHandler[Self]] | None = None,
    ) -> None:
        """
        Intialize a new status bar content producer, i.e. an element.

        The `instance` argument will be provided by swaystatus when the element
        is created. This should not be provided by the subclass.

        The dictionary `env` will be added to the execution environment of any
        click handler. Upon instantiation, the value is a combination of the
        values in configuration for `env`, `settings.<name>.env`, and
        `settings."<name>:<instance>".env`, merged together in that order, with
        later values overriding earlier ones. Additionally, the attributes
        `name` and `instance` (if set) are added.

        The dictionary `on_click` maps pointer button numbers to click handlers
        (i.e. functions or shell commands) which take precedence over any
        already defined on the class.

        When subclassing, there could be more keyword arguments passed,
        depending on its settings in the configuration file.

        To illustrate, let's change the clock example from earlier to allow
        configuration of how the element displays the time. Add an initializer
        that accepts a keyword argument, and change the `blocks` method to use
        the setting:

            >>> from time import strftime
            >>> from typing import Iterator
            >>> from swaystatus import BaseElement, Block
            >>> class Element(BaseElement):
            >>>     def __init__(self, full_text="%c", **kwargs) -> None:
            >>>         super().__init__(**kwargs)
            >>>         self.full_text = full_text
            >>>     def blocks(self) -> Iterator[Block]:
            >>>         yield self.block(strftime(self.full_text))

        Without any further changes, it will behave as it did originally, but
        now it can be configured by adding something like the following to the
        configuration file:

            [settings.clock]
            full_text = "The time here is: %r"

        If there are other instances of the element, they will inherit the
        settings, but they can be overridden:

            [settings."clock:home"]
            full_text = "The time at home is: %r"

            [settings."clock:home".env]
            TZ = 'Asia/Tokyo'
        """
        self.instance = instance
        self.env = env or {}
        if on_click:
            for button, handler in on_click.items():
                self.set_click_handler(button, handler)

    def __str__(self) -> str:
        return f"element name={self.name!r} instance={self.instance!r}"

    def blocks(self) -> Iterator[Block]:
        """
        Yield blocks of content to display on the status bar.

        To create a block, it's recommended to use the `block` method so that
        the block has the proper name and instance set and that there is some
        text:

            >>> from typing import Iterator
            >>> from swaystatus import BaseElement, Block
            >>> class Element(BaseElement):
            >>>     def blocks(self) -> Iterator[Block]:
            >>>         yield self.block("Howdy!")

        See the documentation for `block` for a more detailed explanation.
        """
        raise NotImplementedError

    def block(self, full_text: str, **kwargs) -> Block:
        """
        Create a block of content associated with this element.

        This method ensures that the `Block` it creates is linked to the
        element yielding it, i.e. the name and instance attributes are set
        correctly, which allows click events to be sent to this element.

        One potential issue happens when an element instance has been declared
        in the configuration `order` with the "name:instance" form and the
        element class is also yielding blocks with dynamic instance attributes.
        When swaybar sends click events from these blocks, swaystatus is unable
        to match it with any of the elements it knows about and falls back to
        sending it to the "name" element, which may or may not exist, and is
        definitely not the sender.

        To illustrate the problem, consider an element that yields blocks that
        could be different at every update. For example, there could be an
        element that shows network interfaces:

            >>> from pathlib import Path
            >>> from typing import Iterator
            >>> from swaystatus import BaseElement, Block
            >>> class Element(BaseElement):
            >>>     def blocks(self) -> Iterator[Block]:
            >>>         for dev in Path('/sys/class/net').iterdir():
            >>>             yield self.block(dev.name, instance=dev.name)

        The difference between this and the clock example is that the clock
        instances are not dynamic. Swaystatus knows there will always be two
        instances and knows how to reach them.

        The blocks coming from this new example have potentially different
        instances at every update, so swaystatus can only reach this element by
        the module name.

        The consequence is that an element's instances can either be declared
        statically in the configuration or the blocks created with instances at
        runtime, but not both. Trying to yield a block from the element "foo"
        with its instance set to "a" when the element was already declared in
        the configuration `order` as "foo:b" will mean that click events will
        be lost, or worse, sent to the wrong element.

        Using the `block` method will ensure that the returned `Block` has the
        name and instance set correctly. If the block's instance is set
        dynamically and the element's instance was declared in the
        configuration, an exception will be raised.

        See the documentation for `swaystatus.block` for a full specification
        of `Block`'s available fields which can be passed as keyword arguments
        to this method.
        """
        if self.instance is not None:
            kwargs["instance"] = self.instance
        return Block(name=self.name, full_text=full_text, **kwargs)

    def set_click_handler(self, button: int, handler: ClickHandler[Self]) -> None:
        """
        Add a method to this instance that calls `handler` when blocks from
        this element are clicked with the pointer `button`.

        During execution of the handler, additions from any `env` configuration
        and all attributes of the event will be added to the execution
        environment.

        The `handler` can be one of the following:

            - A shell command compatible with `Popen`, i.e. a string or list of
              strings. The stdout and stderr streams will be logged at the
              `DEBUG` and `ERROR` levels, respectively. It will be allowed to
              finish in a separate thread, and if the return code is zero, the
              status bar will be updated.

            - A function that accepts two positional arguments (this element
              instance and a `ClickEvent`) and returns one of the following:

                  - A shell command. It will be handled like the first item.

                  - A `Popen` object. It will be allowed to finish in a
                    separate thread, and if the return code is zero, the status
                    bar will be updated.

                  - A function that returns a `bool`. It will be run in a
                    separate thread, and if it returns `True`, the status bar
                    will be updated.

                  - A `bool`. If `True`, the status bar will be updated.

                  - `None` and the status bar is not updated.
        """
        method_name = f"on_click_{button}"
        handler_kind = "function" if callable(handler) else "shell command"
        handler_desc = f"{self} {method_name} {handler_kind} handler"

        def handle_shell_command(command: ShellCommand) -> Popen:
            logger.debug(f"executing shell command={command!r}")
            return ShellCommandProcess(command)

        if callable(handler):

            def handler_wrapped(element: Self, click_event: ClickEvent) -> ClickHandlerResult:
                result = handler(element, click_event)
                return handle_shell_command(result) if isinstance(result, (str, list)) else result

        else:

            def handler_wrapped(element: Self, click_event: ClickEvent) -> ClickHandlerResult:
                return handle_shell_command(handler)

        def method_wrapped(self: Self, click_event: ClickEvent) -> ClickHandlerResult:
            logger.debug(f"executing {handler_desc} => {handler}")
            with environ_update(**self.env | click_event.as_dict()):
                try:
                    return handler_wrapped(self, click_event)
                except Exception:
                    logger.exception(f"unhandled exception in {handler_desc}")
            return None

        logger.debug(f"setting {handler_desc} => {handler}")
        setattr(self, method_name, MethodType(method_wrapped, self))

    def on_click(self, click_event: ClickEvent) -> ClickHandlerResult:
        """Delegate a click event to the handler corresponding to its button."""
        try:
            return getattr(self, f"on_click_{click_event.button}")(click_event)
        except AttributeError:
            return None


__all__ = [BaseElement.__name__]



import os
from pathlib import Path

cache_home = Path(os.environ.get("XDG_CACHE_HOME", "~/.cache")).expanduser()
config_home = Path(os.environ.get("XDG_CONFIG_HOME", "~/.config")).expanduser()
data_home = Path(os.environ.get("XDG_DATA_HOME", "~/.local/share")).expanduser()
state_home = Path(os.environ.get("XDG_STATE_HOME", "~/.local/state")).expanduser()
